//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "ProblemNonLin.h"
#include "NonLinSolver.h"
#include "CreatorMap.h"
#include "NonLinUpdater.h"
#include "Traverse.h"
#include "AdaptInfo.h"

namespace AMDiS {

  void ProblemNonLin::initialize(Flag initFlag,
				 ProblemStat *adoptProblem /*= NULL*/,
				 Flag adoptFlag /*= INIT_NOTHING*/) 
  {
    ProblemStat::initialize(initFlag, adoptProblem, adoptFlag);

    // === create Updater ===
    if (updater_) { 
      WARNING("updater already created\n");
    } else {
      if (initFlag.isSet(INIT_UPDATER) || 
	  ((!adoptFlag.isSet(INIT_UPDATER)) && initFlag.isSet(INIT_NONLIN_SOLVER))) {
	createUpdater();
      } 

      if (adoptProblem && 
	  (adoptFlag.isSet(INIT_UPDATER) ||  
	   ((!initFlag.isSet(INIT_UPDATER))&& adoptFlag.isSet(INIT_NONLIN_SOLVER)))) {
	TEST_EXIT(updater_==NULL)("updater already created\n");
	updater_ = dynamic_cast<ProblemNonLin*>(adoptProblem)->getUpdater();
      }
    }

    if (updater_ == NULL) 
      WARNING("no updater created\n");

    // === create nonlinear solver ===
    if (nonLinSolver_) { 
      WARNING("nonlinear solver already created\n");
    } else {
      if (initFlag.isSet(INIT_NONLIN_SOLVER))
	createNonLinSolver();     

      if (adoptProblem && adoptFlag.isSet(INIT_NONLIN_SOLVER)) {
	TEST_EXIT(nonLinSolver_ == NULL)("nonlinear solver already created\n");
	nonLinSolver_ = dynamic_cast<ProblemNonLin*>(adoptProblem)->getNonLinSolver();
      }
    }

    if (nonLinSolver_ == NULL)
      WARNING("no nonlinear solver created\n");
  }


  void ProblemNonLin::createNonLinSolver() 
  {
    // create non-linear solver
    std::string nonLinSolverType("no");
    Parameters::get(name + "->nonlin solver", nonLinSolverType);

    NonLinSolverCreator *nonLinSolverCreator = 
      dynamic_cast<NonLinSolverCreator*>(CreatorMap<NonLinSolver>::getCreator(nonLinSolverType));

    nonLinSolverCreator->setLinearSolver(solver);
    nonLinSolverCreator->setName(name + "->nonlin solver");
    nonLinSolverCreator->setNonLinUpdater(updater_);
    nonLinSolver_ = nonLinSolverCreator->create();
  }


  void ProblemNonLin::solve(AdaptInfo *adaptInfo) 
  {
    TEST_EXIT(nonLinSolver_)("no non-linear solver!\n");
    nonLinSolver_->solve(solverMatrix, *solution, *rhs);
  }


  void ProblemNonLin::buildAfterCoarsen(AdaptInfo *adaptInfo, Flag, bool, bool)
  {
    FUNCNAME("ProblemNonLin::buildAfterCoarsen()");
    int nMeshes = static_cast<int>(meshes.size());

    for (int i = 0; i < nMeshes; i++)
      meshes[i]->dofCompress();    

    for (int i = 0; i < nComponents; i++) {
      MSG("%d DOFs for %s\n", 
	  componentSpaces[i]->getAdmin()->getUsedSize(), 
	  componentSpaces[i]->getName().c_str());
    }
    
    TraverseStack stack;
    ElInfo *elInfo;
    
    // for all elements ...
    for (int i = 0; i < nComponents; i++) {
      elInfo = stack.traverseFirst(componentMeshes[i], -1, 
				   Mesh::CALL_LEAF_EL | 
				   Mesh::FILL_BOUND |
				   Mesh::FILL_COORDS |
				   Mesh::FILL_DET |
				   Mesh::FILL_GRD_LAMBDA);
      while (elInfo) {
	if (solution->getDOFVector(i)->getBoundaryManager()) {
	  solution->getDOFVector(i)->
	    getBoundaryManager()->fillBoundaryConditions(elInfo, solution->getDOFVector(i));
	}
	elInfo = stack.traverseNext(elInfo);
      }
    }
  }

}
