// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file ProblemNonLin.h */

#ifndef AMDIS_PROBLEMNONLIN_H_
#define AMDIS_PROBLEMNONLIN_H_

#include "ProblemStat.h"
#include "NonLinUpdater.h"
#include "NonLinSolver.h"
#include "DOFVector.h"
#include "SystemVector.h"
#include "MatrixVector.h"

namespace AMDiS {

  /**
   * \ingroup Problem
   * 
   * \brief
   * Standard implementation for a vector valued non linear problem.
   */
  class ProblemNonLin : public ProblemStat
  {
  public:
    /// Constructs a ProblemNonLin with given name.
    ProblemNonLin(const std::string& name_)  
      : ProblemStat(name_.c_str()),
	nonLinSolver_(NULL),
	updater_(NULL)
    {
      u0_.resize(nComponents);
      u0_.set(NULL);
    }

    /** \brief
     * Sets \ref u0_ and interpolates it to \ref solution_.
     */
    inline void setU0(AbstractFunction<double, WorldVector<double> > *u0Fct,
		      int index) 
    {
      FUNCNAME("ProblemNonLinVec::setU0()");
      TEST_EXIT(index < nComponents)("invalid index\n");
      u0_[index] = u0Fct;
      solution->getDOFVector(index)->interpol(u0Fct);
    }

    /** \brief
     * Destructor.
     */
    virtual ~ProblemNonLin() {}

    /** \brief
     * Initialisation of the problem.
     */
    virtual void initialize(Flag initFlag,
			    ProblemStat *adoptProblem = NULL,
			    Flag adoptFlag = INIT_NOTHING);

    /** \brief
     * Used in \ref initialize().
     */
    virtual void createUpdater() 
    {
      updater_ = new NonLinUpdater(systemMatrix);
    }

    /** \brief
     * Used in \ref initialize().
     */
    virtual void createNonLinSolver();

    /** \brief
     * Overrides ProblemVec::solve(). Uses the non linear solver 
     * \ref nonLinSolver_.
     */
    virtual void solve(AdaptInfo *adaptInfo);

    /** \brief
     * Overrides Problem::buildAfterCoarsen(). Sets dirichlet boundaries
     * in \ref A_ and \ref rhs_.
     */
    virtual void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag,
				   bool assembleMatrix = true,
				   bool assembleVector = true);

    /** \brief
     * Returns /ref nonLinSolver_.
     */
    inline NonLinSolver *getNonLinSolver() 
    {
      return nonLinSolver_;
    }

    /** \brief
     * Returns \ref updater.
     */
    inline NonLinUpdater* getUpdater() 
    {
      return updater_;
    }

    /** \brief
     * Sets \ref nonLinSolver_.
     */
    inline void setNonLinSolver(NonLinSolver *nonLinSolver) 
    {
      nonLinSolver_ = nonLinSolver;
    }

    /** \brief
     * Sets \ref updater.
     */
    inline void setUpdater(NonLinUpdater *updater) 
    {
      updater_ = updater;
    }


  protected:
    /** \brief
     * Initial guess for the solution.
     */
    Vector<AbstractFunction<double, WorldVector<double> >*> u0_;

    /** \brief
     * Non linear solver. Used in \ref solve().
     */
    NonLinSolver *nonLinSolver_;

    /** \brief
     * Updator
     */
    NonLinUpdater *updater_;
  };

}

#endif

